from __future__ import annotations

import os
import shlex
from collections.abc import Iterable, Sequence
from functools import wraps
from pathlib import Path
from typing import Any, Callable, TypeVar, cast

T = TypeVar("T")
SelfType = TypeVar("SelfType")


def prepare_mkvtoolnix_path(
    path: str | os.PathLike | Iterable[str],
) -> tuple[str, ...]:
    """
    Parameters
    ----------
    path : str | list[str] | os.PathLike | tuple
        The path to prepare for use with MKVToolNix.

    Returns
    -------
    tuple[str, ...]
        The prepared path as a list of strings.

    Raises
    ------
    ValueError
        If the path type is invalid. Expected str, list of str, or os.PathLike.
    """
    if isinstance(path, os.PathLike):
        return (os.fspath(path),)
    elif isinstance(path, str):  # noqa: RET505
        # Check if the path exists and is accessible
        return (path,) if Path(path).exists() else tuple(shlex.split(path))
    elif isinstance(path, tuple):
        return path
    elif isinstance(path, Sequence):
        return tuple(path)
    else:
        msg = "Invalid path type. Expected str, List[str], Tuple[str] or os.PathLike."
        raise ValueError(msg)  # noqa: TRY004


def ensure_info(
    info_attr: str,
    fetch_func: Callable[..., T],
    fetch_args: list[str],
    **fetch_kwargs: Any,  # noqa: ANN401
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    """
    A decorator that ensures the specified attribute (info_attr) is available before executing the method.

    If the attribute doesn't exist or is None, it fetches the information using the provided fetch function.

    Parameters:
    -----------
    info_attr : str
        The name of the attribute to check and potentially fetch.
    fetch_func : Callable
        The function to call to fetch the information if it's not available.
    fetch_args : list[str]
        A list of argument names to pass to fetch_func. These can be attributes of the decorated object or
         literal strings.
    **fetch_kwargs : Any
        Additional keyword arguments to pass to fetch_func.

    Returns:
    --------
    Callable
        A decorator function that wraps the original method.

    Example:
    --------
    @ensure_info('_info', get_info, ['file_path', 'mkvmerge_path'])
    def some_method(self):
        # Use self._info here
        pass
    """

    def decorator(method: Callable[..., T]) -> Callable[..., T]:
        @wraps(method)
        def wrapper(
            self: SelfType,
            *args: list[Any],
            **kwargs: dict[str, Any],
        ) -> T:
            if not hasattr(self, info_attr) or getattr(self, info_attr) is None:
                func_params = []
                for arg in fetch_args:
                    if hasattr(self, arg):
                        func_params.append(getattr(self, arg))
                    else:
                        func_params.append(arg)
                setattr(self, info_attr, fetch_func(*func_params, **fetch_kwargs))
            return method(self, *args, **kwargs)

        return cast(Callable[..., T], wrapper)

    return decorator



"""
:class:`~pymkv.MKVAttachment` classes are used to represent attachment files within an MKV or to be used in an
MKV.

Examples
--------
Below are some basic examples of how the :class:`~pymkv.MKVAttachment` objects can be used.

Create a new :class:`~pymkv.MKVAttachment` and add it to an :class:`~pymkv.MKVFile`.

>>> from pymkv import MKVAttachment
>>> attachment = MKVAttachment('path/to/attachment.jpg', name='NAME')
>>> attachment.description = 'DESCRIPTION'

Attachments can also be added directly to an :class:`~pymkv.MKVFile`.

>>> from pymkv import MKVFile
>>> mkv = MKVFile('path/to/file.mkv')
>>> mkv.add_attachment('path/to/other/attachment.png')

Now, the MKV can be muxed with both attachments.

>>> mkv.add_attachment(attachment)
>>> mkv.mux('path/to/output.mkv')
"""

from __future__ import annotations

from mimetypes import guess_type
from pathlib import Path


class MKVAttachment:
    """A class that represents an MKV attachment for an :class:`~pymkv.MKVFile` object.

    Parameters
    ----------
    file_path : str
        The path to the attachment file.
    name : str, optional
        The name that will be given to the attachment when muxed into a file.
    description : str, optional
        The description that will be given to the attachment when muxed into a file.
    attach_once : bool, optional
        Determines if the attachment should be added to all split files or only the first. Default is False,
        which will attach to all files.

    Attributes
    ----------
    mime_type : str
        The attachment's MIME type. The type will be guessed when :attr:`~pymkv.MKVAttachment.file_path` is set.
    name : str
        The name that will be given to the attachment when muxed into a file.
    description : str
        The description that will be given to the attachment when muxed into a file.
    attach_once : bool
        Determines if the attachment should be added to all split files or only the first. Default is False,
        which will attach to all files.
    """

    def __init__(
        self,
        file_path: str,
        name: str | None = None,
        description: str | None = None,
        attach_once: bool | None = False,
    ) -> None:
        self.mime_type: str | None = None
        self._file_path: str
        self.file_path = file_path
        self.name = name
        self.description = description
        self.attach_once = attach_once

    def __repr__(self) -> str:
        """
        Return a string representation of the object.

        Parameters:
            self (object): The object for which the string representation is generated.

        Returns:
            str: The string representation of the object. It is the representation of the object's __dict__ attribute.
        """
        return repr(self.__dict__)

    @property
    def file_path(self) -> str:
        """str: The path to the attachment file.

        Raises
        ------
        FileNotFoundError
            Raised if `file_path` does not exist.
        """
        return self._file_path

    @file_path.setter
    def file_path(self, file_path: str) -> None:
        """
        Parameters
        ----------
        file_path : str
            The file path to be set.

        Raises
        ------
        FileNotFoundError
            If the specified file does not exist.

        Returns
        -------
        None
        """
        fp = Path(file_path).expanduser()
        if not fp.is_file():
            msg = f'"{fp}" does not exist'
            raise FileNotFoundError(msg)
        self.mime_type = guess_type(fp)[0]
        self.name = None
        self._file_path = str(fp)



from __future__ import annotations

import json
import os
import subprocess as sp
from collections.abc import Iterable, Sequence
from pathlib import Path
from re import match
from typing import Any

from pymkv.utils import prepare_mkvtoolnix_path


def checking_file_path(file_path: str | os.PathLike[Any] | None) -> str:
    """
    Parameters
    ----------
    file_path : str, os.PathLike
        The path to the file that needs to be checked.

    Returns
    -------
    str
        The resolved and expanded file path if it exists.

    Raises
    ------
    TypeError
        If the provided file path is not of type str.
    FileNotFoundError
        If the file path does not exist.
    """
    if not isinstance(file_path, (str, os.PathLike)):
        msg = f'"{file_path}" is not of type str'
        raise TypeError(msg)
    file_path = Path(file_path).expanduser()
    if not file_path.is_file():
        msg = f'"{file_path}" does not exist'
        raise FileNotFoundError(msg)
    return str(file_path)


def get_file_info(
    file_path: str | os.PathLike[Any],
    mkvmerge_path: str | os.PathLike | Iterable[str],
    check_path: bool = True,
) -> dict[str, Any]:
    """
    Get information about a media file using mkvmerge.

    Parameters:
    -----------
    file_path : Union[str, os.PathLike]
        The path to the media file to analyze.
    mkvmerge_path : Union[str, tuple, os.PathLike]
        The path to the mkvmerge executable or a list of command parts.
    check_path : bool, optional
        Whether to check and validate the file path. Defaults to True.

    Returns:
    --------
    Dict
        A dictionary containing the parsed JSON output from mkvmerge,
        which includes detailed information about the media file.

    Raises:
    -------
    subprocess.CalledProcessError
        If mkvmerge fails to execute or returns a non-zero exit status.
    json.JSONDecodeError
        If the output from mkvmerge cannot be parsed as JSON.
    FileNotFoundError
        If check_path is True and the file does not exist.
    TypeError
        If check_path is True and file_path is not a string or PathLike object.
    """
    if check_path:
        file_path = checking_file_path(file_path)

    cmds = [*prepare_mkvtoolnix_path(mkvmerge_path), "-J", file_path]
    return json.loads(sp.check_output(cmds).decode())  # noqa: S603


def verify_mkvmerge(
    mkvmerge_path: str | os.PathLike | Iterable[str] = "mkvmerge",
) -> bool:
    """
    Parameters
    ----------
    mkvmerge_path : str | None, optional
        The path to the `mkvmerge` executable. If not provided, the default value is "mkvmerge".

    Returns
    -------
    bool
        True, if `mkvmerge_path` is valid and the `mkvmerge` executable is found. False otherwise.
    """
    try:
        mkvmerge_command = list(prepare_mkvtoolnix_path(mkvmerge_path))
        mkvmerge_command.append("-V")
        output = sp.check_output(mkvmerge_command).decode()  # noqa: S603
    except (sp.CalledProcessError, FileNotFoundError):
        return False
    return bool(match("mkvmerge.*", output))


def verify_matroska(
    file_path: str | os.PathLike[Any],
    mkvmerge_path: str | os.PathLike | Iterable[str] = "mkvmerge",
) -> bool:
    """
    Parameters
    ----------
    file_path : str or os.PathLike
        The path to the Matroska file to be verified.

    mkvmerge_path : str, optional
        The path to the `mkvmerge` executable. Default is "mkvmerge".

    Returns
    -------
    bool
        True if the Matroska file is valid and is of type "Matroska", False otherwise.

    Raises
    ------
    FileNotFoundError
        If `mkvmerge` executable is not found at the specified path.
    TypeError
        If `file_path` is not a string or an instance of `os.PathLike`.
    FileNotFoundError
        If the specified `file_path` does not exist.
    ValueError
        If the `file_path` could not be opened using `mkvmerge`.

    Notes
    -----
    This method verifies the validity of a Matroska file by checking if it is of type "Matroska"
    using the `mkvmerge` command-line tool.
    """
    if not verify_mkvmerge(mkvmerge_path=mkvmerge_path):
        msg = "mkvmerge is not at the specified path, add it there or change the mkvmerge_path property"
        raise FileNotFoundError(msg)
    try:
        if isinstance(mkvmerge_path, Sequence) and not isinstance(mkvmerge_path, str):
            mkvmerge_path = tuple(mkvmerge_path)
        info_json: dict[str, Any] = get_file_info(file_path, mkvmerge_path)

    except sp.CalledProcessError as e:
        msg = f'"{file_path}" could not be opened'
        raise ValueError(msg) from e
    return info_json["container"]["type"] == "Matroska"


def verify_file_path_and_mkvmerge(
    file_path: str | os.PathLike[Any],
    mkvmerge_path: str | os.PathLike | Iterable[str] = "mkvmerge",
) -> str:
    """
    Parameters
    ----------
    file_path : Union[str, os.PathLike]
        The path to the file that needs to be verified.

    mkvmerge_path : str, optional
        The path to the mkvmerge executable. By default, it is set to "mkvmerge".

    Returns
    -------
    str
        The verified file path.

    Raises
    ------
    FileNotFoundError
        If mkvmerge is not found at the specified path or if the file_path does not exist.

    TypeError
        If the file_path is not of type str.

    """
    if not verify_mkvmerge(mkvmerge_path=mkvmerge_path):
        msg = "mkvmerge is not at the specified path, add it there or change the mkvmerge_path property"
        raise FileNotFoundError(msg)
    return checking_file_path(file_path)


def verify_recognized(
    file_path: str | os.PathLike[Any],
    mkvmerge_path: str | os.PathLike | Iterable[str] = "mkvmerge",
) -> bool:
    """
    Parameters
    ----------
    file_path : Union[str, os.PathLike]
        The path to the file that will be verified.

    mkvmerge_path : str, optional
        The path to the `mkvmerge` executable. Default is "mkvmerge".

    Returns
    -------
    bool
        True if the container format of the file is recognized, False otherwise.
    """
    file_path = verify_file_path_and_mkvmerge(file_path, mkvmerge_path)
    try:
        if isinstance(mkvmerge_path, list):
            mkvmerge_path = tuple(mkvmerge_path)
        info_json = get_file_info(file_path, mkvmerge_path, check_path=False)
    except sp.CalledProcessError as e:
        msg = f'"{file_path}" could not be opened'
        raise ValueError(msg) from e
    return info_json["container"]["recognized"]


def verify_supported(  # noqa: ANN201
    file_path: str | os.PathLike[Any],
    mkvmerge_path: str | os.PathLike | Iterable[str] = "mkvmerge",
):
    """
    Parameters
    ----------
    file_path : Union[str, os.PathLike]
        The path of the file to be verified.

    mkvmerge_path : str, optional
        The path to the `mkvmerge` executable. Defaults to "mkvmerge".

    Returns
    -------
    bool
        Whether the container format of the file is supported by `mkvmerge`.

    Raises
    ------
    ValueError
        If the file cannot be opened or an error occurs during the verification process.
    """
    mkvmerge_path = prepare_mkvtoolnix_path(mkvmerge_path)
    file_path = verify_file_path_and_mkvmerge(file_path, mkvmerge_path)
    try:
        if isinstance(mkvmerge_path, list):
            mkvmerge_path = tuple(mkvmerge_path)
        info_json = get_file_info(file_path, mkvmerge_path, check_path=False)
    except sp.CalledProcessError as e:
        msg = '"{}" could not be opened'
        raise ValueError(msg) from e
    return info_json["container"]["supported"]



from __future__ import annotations

import importlib.metadata

# package imports
from .MKVAttachment import MKVAttachment
from .MKVFile import MKVFile
from .MKVTrack import MKVTrack
from .Timestamp import Timestamp
from .Verifications import (
    checking_file_path,
    get_file_info,
    verify_matroska,
    verify_mkvmerge,
    verify_recognized,
    verify_supported,
)

# set the version number within the package using importlib
try:
    __version__: str | None = importlib.metadata.version("pymkv2")
except importlib.metadata.PackageNotFoundError:
    # package is not installed
    __version__ = None

__all__ = (
    "MKVAttachment",
    "MKVFile",
    "MKVTrack",
    "Timestamp",
    "verify_matroska",
    "verify_mkvmerge",
    "verify_recognized",
    "verify_supported",
    "checking_file_path",
    "get_file_info",
)



"""
:class:`~pymkv.MKVTrack` classes are used to represent tracks within an MKV or to be used in an MKV. They can
represent a video, audio, or subtitle track.

Examples
--------

Below are some basic examples of how the :class:`~pymkv.MKVTrack` objects can be used.

Create a new :class:`~pymkv.MKVTrack` from a track file. This example takes a standalone track file and uses it in an
:class:`~pymkv.MKVTrack`.

>> from pymkv import MKVTrack
>> track1 = MKVTrack("path/to/track.h264")
>> track1.track_name = "Some Name"
>> track1.language = "eng"

Create a new :class:`~pymkv.MKVTrack` from an MKV file. This example will take a specific track from an MKV and also
prevent any global tags from being included if the :class:`~pymkv.MKVTrack` is muxed into an :class:`~pymkv.MKVFile`.

>> track2 = MKVTrack("path/to/track.aac")
>> track2.language = "eng"

Create a new :class:`~pymkv.MKVTrack` from an MKV file. This example will take a specific track from an MKV and also
prevent any global tags from being included if the :class:`~pymkv.MKVTrack` is muxed into an :class:`~pymkv.MKVFile`.

>> track3 = MKVTrack("path/to/MKV.mkv", track_id=1)
>> track3.no_global_tags = True

Now all these tracks can be added to an :class:`~pymkv.MKVFile` object and muxed together.

>> from pymkv import MKVFile
>> file = MKVFile()
>> file.add_track(track1)
>> file.add_track(track2)
>> file.add_track(track3)
>> file.mux("path/to/output.mkv")
"""

from __future__ import annotations

import os
import subprocess as sp
from pathlib import Path
from typing import TYPE_CHECKING, Any

from pymkv.ISO639_2 import is_iso639_2
from pymkv.utils import prepare_mkvtoolnix_path
from pymkv.Verifications import checking_file_path, get_file_info, verify_supported

if TYPE_CHECKING:
    from collections.abc import Iterable


class MKVTrack:
    """A class that represents a track for an :class:`~pymkv.MKVFile` object.
    :class:`~pymkv.MKVTrack` objects are video, audio, or subtitles. Tracks can be standalone files or a single track
    within an MKV file, both can be handled by pymkv. An :class:`~pymkv.MKVTrack` object can be added to an
    :class:`~pymkv.MKVFile` and will be included when the MKV is muxed.
    Parameters
    ----------
    file_path : str
        Path to the track file. This can also be an MKV where the `track_id` is the track represented in the MKV.
    track_id : int, optional
        The id of the track to be used from the file. `track_id` only needs to be set when importing a track from
        an MKV. In this case, you can specify `track_id` to indicate which track from the MKV should be used. If not
        set, it will import the first track. Track 0 is imported by default because mkvmerge sees standalone
        track files as having one track with track_id set as 0.
    track_name : str, optional
        The name that will be given to the track when muxed into a file.
    language : str, optional
        The language of the track. It must be an ISO639-2 language code.
    language_ietf : str, optional
        The language of the track. It must be a BCP47 language code. Has priority over 'language'.
    default_track : bool, optional
        Determines if the track should be the default track of its type when muxed into an MKV file.
    forced_track : bool, optional
        Determines if the track should be a forced track when muxed into an MKV file.
    mkvmerge_path : str, list, os.PathLike, optional
        The path where pymkv looks for the mkvmerge executable. pymkv relies on the mkvmerge executable to parse
        files. By default, it is assumed mkvmerge is in your shell's $PATH variable. If it is not, you need to set
        *mkvmerge_path* to the executable location.
    mkvextract_path : str, list, os.PathLike, optional
        The path where pymkv looks for the mkvextract executable. pymkv relies on the mkvextract executable to extract
        files. By default, it is assumed mkvextract is in your shell's $PATH variable. If it is not, you need to set
        *mkvextract_path* to the executable location.
    tag_entries : int, optional
        The number of tag entries.
    Attributes
    ----------
    mkvmerge_path : list
        The path of the mkvmerge executable.
    mkvextract_path : list
        The path of the mkvextract executable.
    track_name : str
        The name that will be given to the track when muxed into a file.
    default_track : bool
        Determines if the track should be the default track of its type when muxed into an MKV file.
    forced_track : bool
        Determines if the track should be a forced track when muxed into an MKV file.
    no_chapters : bool
        If chapters exist in the track file, don't include them when this :class:`~pymkv.MKVTrack` object is a track
        in an :class:`~pymkv.MKVFile` mux operation. This option has no effect on standalone track files, only tracks
        that are already part of an MKV file.
    no_global_tags : bool
        If global tags exist in the track file, don't include them when this :class:`~pymkv.MKVTrack` object is a track
        in an :class:`~pymkv.MKVFile` mux operation. This option has no effect on standalone track files, only tracks
        that are already part of an MKV file.
    no_track_tags : bool
        If track tags exist in the specified track within the track file, don't include them when this
        :class:`~pymkv.MKVTrack` object is a track in an :class:`~pymkv.MKVFile` mux operation. This option has no
        effect on standalone track files, only tracks that are already part of an MKV file.
    no_attachments : bool
        If attachments exist in the track file, don't include them when this :class:`~pymkv.MKVTrack` object is a track
        in an :class:`~pymkv.MKVFile` mux operation. This option has no effect on standalone track files, only tracks
        that are already part of an MKV file.
    """

    def __init__(  # noqa: PLR0913
        self,
        file_path: str,
        track_id: int = 0,
        track_name: str | None = None,
        language: str | None = None,
        language_ietf: str | None = None,
        default_track: bool | None = False,
        forced_track: bool | None = False,
        flag_commentary: bool | None = False,
        flag_hearing_impaired: bool | None = False,
        flag_visual_impaired: bool | None = False,
        flag_original: bool | None = False,
        mkvmerge_path: str | os.PathLike | Iterable[str] = "mkvmerge",
        mkvextract_path: str | os.PathLike | Iterable[str] = "mkvextract",
        sync: int | None = None,
        existing_info: dict[str, Any] | None = None,
        tag_entries: int = 0,
    ) -> None:
        from pymkv.TypeTrack import get_track_extension

        # track info
        self._track_codec: str | None = None
        self._track_type: str | None = None

        # base
        self.mkvmerge_path = prepare_mkvtoolnix_path(mkvmerge_path)
        self._info_json: dict[str, Any] | None = existing_info or None
        self._file_path: str
        self.file_path = file_path
        self._track_id: int
        self.track_id = track_id
        self._file_id = 0
        self._pts = 0

        # flags
        self.track_name = track_name
        self._language: str | None = None
        self.language = language
        self._sync: int | None = None
        self.sync = sync
        self._language_ietf: str | None = None
        self.language_ietf = language_ietf
        self._tags: str | None = None
        self.default_track = default_track
        self.forced_track = forced_track
        self.flag_commentary = flag_commentary
        self.flag_hearing_impaired = flag_hearing_impaired
        self.flag_visual_impaired = flag_visual_impaired
        self.flag_original = flag_original
        self._tag_entries = tag_entries

        # exclusions
        self.no_chapters = False
        self.no_global_tags = False
        self.no_track_tags = False
        self.no_attachments = False

        # mkvextract
        self.mkvextract_path = prepare_mkvtoolnix_path(mkvextract_path)
        self.extension = get_track_extension(self)

    def __repr__(self) -> str:
        """
        Return a string representation of the MKVTrack object.

        Returns:
            str: A string representation of the object's dictionary.
        """
        return repr(self.__dict__)

    @property
    def file_path(self) -> str:
        """
        Get the path to the track or MKV file containing the desired track.

        Returns:
            str: The file path.

        Raises:
            ValueError: If `file_path` is not a supported file type.
        """
        return self._file_path

    @file_path.setter
    def file_path(self, file_path: str) -> None:
        """
        Set the file path for the track.

        This method checks if the provided file path is valid and supported by mkvmerge.
        If the file is valid, it sets the file path and resets the track_id to 0.

        Args:
            file_path (str): The path to the file containing the track.

        Raises:
            ValueError: If the file is not a valid Matroska file or is not supported.
        """
        file_path = checking_file_path(file_path)
        if not verify_supported(file_path, mkvmerge_path=self.mkvmerge_path):
            msg = f"The file '{file_path}' is not a valid Matroska file or is not supported."
            raise ValueError(msg)
        self._file_path = file_path
        self.track_id = 0

    @property
    def file_id(self) -> int:
        """
        Get the ID of the file the track belongs to.

        The file ID represents which file the current track is associated with. This is particularly useful
        when handling multiple files.

        Returns:
            int: The file ID.

        Raises:
            IndexError: If the passed in index is out of range of the file's tracks.
        """
        return self._file_id

    @file_id.setter
    def file_id(self, file_id: int) -> None:
        """
        Set the ID of the file the track belongs to.

        Args:
            file_id (int): The file ID to set.

        Raises:
            ValueError: If file_id is not an integer.
        """
        if isinstance(file_id, int):
            self._file_id = file_id
        else:
            msg = "file_id must be an integer"
            raise ValueError(msg)  # noqa: TRY004

    @property
    def track_id(self) -> int:
        """
        Get the ID of the track within the file.

        Returns:
            int: The track ID.

        Raises:
            IndexError: If the passed in index is out of range of the file's tracks.
        """
        return self._track_id

    @track_id.setter
    def track_id(self, track_id: int) -> None:
        """
        Set the ID of the track within the file.

        This method checks that the ID passed in exists in the file. It then looks at the new track
        and sets the codec and track type. Should be left at 0 unless extracting a specific track from an MKV.

        Args:
            track_id (int): The track ID to set.

        Raises:
            IndexError: If the passed in index is out of range of the file's tracks.
        """
        if not self._info_json:
            self._info_json = get_file_info(self.file_path, mkvmerge_path=self.mkvmerge_path)
        tracks = self._info_json.get("tracks", [])
        if not 0 <= track_id < len(tracks):
            msg = "track index out of range"
            raise IndexError(msg)
        self._track_id = track_id
        try:
            self._pts = tracks[track_id]["start_pts"]
        except KeyError:
            self._pts = 0
        self._track_codec = tracks[track_id]["codec"]
        self._track_type = tracks[track_id]["type"]

    @property
    def language(self) -> str | None:
        """
        Get the language of the track.

        Returns:
            str: The language of the track.

        Raises:
            ValueError: If the passed in language is not an ISO 639-2 language code.
        """
        return self._language

    @language.setter
    def language(self, language: str | None) -> None:
        """
        Set the language of the MKVTrack.

        Args:
            language (str): The language to be set for the MKVTrack.

        Raises:
            ValueError: If the provided language is not a valid ISO639-2 language code.
        """
        if language is None or is_iso639_2(language):
            self._language = language
        else:
            msg = "not an ISO639-2 language code"
            raise ValueError(msg)

    @property
    def pts(self) -> int:
        """
        Get the Presentation Timestamp (PTS) of the track.

        The PTS in multimedia files indicates the exact time when a frame or audio sample
        should be presented to the user, ensuring accurate synchronization between audio and video streams.

        Returns:
            int: The PTS value.
        """
        return self._pts

    @property
    def sync(self) -> int | None:
        """
        Get the synchronization offset for the track.

        This property represents the synchronization offset for the track.
        Positive values delay the track, while negative values advance it.

        Returns:
            int: The current synchronization offset in milliseconds.

        Note:
            Setting this property allows you to adjust the track timing:
            - Positive values delay the track (shift it later in time)
            - Negative values advance the track (shift it earlier in time)

        Example:
            track.sync = 1000  # Delay the track by 1 second
            track.sync = -500  # Advance the track by 0.5 seconds
        """
        return self._sync

    @sync.setter
    def sync(self, sync: int | None) -> None:
        """
        Set the synchronization offset for the track.

        Args:
            sync (int): The synchronization offset in milliseconds.
        """
        self._sync = sync

    @property
    def language_ietf(self) -> str | None:
        """
        Get the language of the track with BCP47 format.

        Returns:
            str: The language of the track in BCP47 format.

        Raises:
            ValueError: If the passed in language is not a BCP47 language code.
        """
        return self._language_ietf

    @language_ietf.setter
    def language_ietf(self, language_ietf: str | None) -> None:
        """
        Set the language of the track with BCP47 format.

        Args:
            language_ietf (str): The language to set in BCP47 format.
        """
        self._language_ietf = language_ietf

    @property
    def tags(self) -> str | None:
        """
        Get the tags file to include with the track.

        Returns:
            str: The path to the tags file.

        Raises:
            FileNotFoundError: If the passed in file does not exist or is not a file.
            TypeError: If the passed in file is not of type str.
        """
        return self._tags

    @tags.setter
    def tags(self, file_path: str | os.PathLike | None) -> None:
        """
        Set the tags file for the track.

        Args:
            file_path (str): The path to the tags file.

        Raises:
            TypeError: If the file_path is not a string.
            FileNotFoundError: If the file does not exist or is not a file.
        """
        if file_path is None:
            self._tags = None
            return

        if not isinstance(file_path, str):
            msg = f'"{file_path}" is not of type str'
            raise TypeError(msg)

        file_path = Path(file_path).expanduser()
        if not file_path.is_file():
            msg = f'"{file_path}" does not exist'
            raise FileNotFoundError(msg)
        self._tags = str(file_path)

    @property
    def tag_entries(self) -> int:
        """
        Gets the number of existing tag entries in the track.

        Returns:
            int: The number of entries.
        """
        return self._tag_entries

    @property
    def track_codec(self) -> str | None:
        """
        Get the codec of the track.

        Returns:
            str: The codec of the track, such as h264 or AAC.
        """
        return self._track_codec

    @property
    def track_type(self) -> str | None:
        """
        Get the type of the track.

        Returns:
            str: The type of track, such as video or audio.
        """
        return self._track_type

    def extract(
        self,
        output_path: str | os.PathLike | None = None,
        silent: bool | None = False,
    ) -> str:
        """
        Extract the track as a file.

        Args:
            output_path (str | os.PathLike | None, optional): The path to be used as the output file
            in the mkvextract command.
            silent (bool | None, optional): By default the mkvmerge output will be shown unless silent is True.

        Returns:
            str: The path of the extracted file.
        """
        extract_info_file = f"_[{self.track_id}]"
        if self.language:
            extract_info_file += f"_{self.language}"
        if self.extension:
            extract_info_file += f".{self.extension}"
        # TODO: Should this be extension?
        if (not self.language and not self.extension) and self.track_name:
            extract_info_file += f"_{self.track_name}"
        if output_path is None:
            output_path = f"{self.file_path}{extract_info_file}"
        else:
            file = Path(self.file_path)
            output_path = Path(output_path, f"{file.name}{extract_info_file}")
        output_path = str(Path(output_path).expanduser())
        command = [
            *self.mkvextract_path,
            "tracks",
            f"{self.file_path}",
            f"{self.track_id}:{output_path}",
        ]
        if silent:
            sp.run(  # noqa: S603
                command,
                stdout=sp.DEVNULL,
                check=True,
            )
        else:
            sp.run(command, check=True, capture_output=True)  # noqa: S603
        return output_path



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pymkv.MKVTrack import MKVTrack

type_files = {
    "video": {
        "V_MPEG1": "mpg",
        "V_MPEG2": "mpg",
        "V_MPEG4/ISO/AVC": "264",
        "MPEG-4p10": "h264",
        "HEVC": "h265",
        "V_MS/VFW/FOURCC": "avi",
        "V_REAL": "rm",
        "V_THEORA": "ogg",
        "V_VP8": "ivf",
        "V_VP9": "ivf",
        "AVC/H.264/MPEG-4p10": "mp4",
    },
    "audio": {
        "AAC": "aac",
        "AC3": "ac3",
        "AC-3": "ac3",
        "ALAC": "caf",
        "DTS": "dts",
        "FLAC": "flac",
        "MPEG/L2": "mp2",
        "MPEG/L3": "mp3",
        "OPUS": "ogg",
        "PCM": "wav",
        "REAL": "ra",
        "TRUEHD": "thd",
        "MLP": "mlp",
        "TTA1": "tta",
        "VORBIS": "ogg",
        "WAVPACK4": "wv",
        "V_MS/VFW/FOURCC, WVC1": "wvc",
        "VC-1": "wvc",
        "Vorbis": "ogg",
    },
    "subtitles": {
        "PGS": "sup",
        "ASS": "ass",
        "SSA": "ssa",
        "UTF8": "srt",
        "SubRip/SRT": "srt",
        "ASCII": "srt",
        "VOBSUB": "sub",
        "USF": "usf",
        "WEBVTT": "vtt",
    },
}


def get_track_extension(track: MKVTrack) -> str | None:
    """
    Get the file extension for a given MKVTrack.

    Parameters
    ----------
    track : MKVTrack
        The MKVTrack object from which to determine the file extension.

    Returns
    -------
    str | None
        The file extension corresponding to the track's type and codec, or None if not found.

    Notes
    -----
    This function uses the track's type and codec to look up the appropriate
    file extension in the `type_files` dictionary. If no matching extension
    is found, it returns None.
    """
    track_type = track.track_type
    track_codec = track.track_codec
    if track_type in type_files and track_codec in type_files[track_type]:
        return type_files[track_type][track_codec]
    return None



from __future__ import annotations

import re
from typing import Final, cast

_TIMESTAMP_PATTERN: Final[re.Pattern] = re.compile(
    r"^[0-9]{1,2}(:[0-9]{1,2}){1,2}(\.[0-9]{1,9})?$",
)


class Timestamp:
    __slots__ = ["_hh", "_mm", "_ss", "_nn", "_form"]

    def __init__(  # noqa: PLR0913
        self,
        timestamp: str | int | Timestamp | None = None,
        hh: int | None = None,
        mm: int | None = None,
        ss: int | None = None,
        nn: int | None = None,
        form: str = "MM:SS",
    ) -> None:
        """
        A class that represents a timestamp used in MKVFiles.

        The Timestamp class represents a timestamp used in mkvmerge. These are commonly used for splitting MKVFiles.
        Specific time values can overridden in the timestamp using 'hh', 'mm', 'ss', and 'nn'. Any override value
        that is greater than its maximum (ex. 61 minutes) will be set to 0.

        timestamp (str, int, Timestamp):
            A str of a timestamp acceptable to mkvmerge or an int representing seconds. This value will be
            the basis of the timestamp.
        hh (int):
            Hours in the timestamp. This will override the hours in the given timestamp.
        mm (int):
            Minutes in the timestamp. This will override the minutes in the given timestamp.
        ss (int):
            Seconds in the timestamp. This will override the seconds in the given timestamp.
        nn (int):
            Nanoseconds in the timestamp. This will override the nanoseconds in the given timestamp.
        form (str):
            A str for the form of the returned timestamp. 'MM' and 'SS' must be included where 'HH' and 'NN' are
            optional but will be included if 'hh' and 'nn' are not zero.
        """
        self._hh = hh
        self._mm = mm
        self._ss = ss
        self._nn = nn
        self._form = form
        if isinstance(timestamp, Timestamp):
            self._hh = timestamp.hh
            self._mm = timestamp.mm
            self._ss = timestamp.ss
            self._nn = timestamp.nn
        elif timestamp is not None:
            self.extract(timestamp)
        else:
            self._hh = 0 if hh is None else hh
            self._mm = 0 if mm is None else mm
            self._ss = 0 if ss is None else ss
            self._nn = 0 if nn is None else nn

    def __eq__(self, other: object) -> bool:
        """
        Compares the Timestamp object with another Timestamp object for equality.

        Args:
            other: The Timestamp object to compare with.

        Returns:
            bool: True if the Timestamp objects are equal, False otherwise.
        """
        if not isinstance(other, Timestamp):
            return NotImplemented

        return self.hh == other.hh and self.mm == other.mm and self.ss == other.ss and self.nn == other.nn

    def __ne__(self, other: object) -> bool:
        """
        Compares the Timestamp object with another Timestamp object for inequality.

        Args:
            other: The Timestamp object to compare with.

        Returns:
            bool: True if the Timestamp objects are not equal, False otherwise.
        """
        if not isinstance(other, Timestamp):
            return NotImplemented
        return self.hh != other.hh or self.mm != other.mm or self.ss != other.ss or self.nn != other.nn

    def __lt__(self, other: object) -> bool:
        """
        Compares the Timestamp object with another Timestamp object to determine if it is less than.

        Args:
            other: The Timestamp object to compare with.

        Returns:
            bool: True if the Timestamp object is less than the other Timestamp object, False otherwise.
        """
        if not isinstance(other, Timestamp):
            return NotImplemented
        if self.hh != other.hh:
            return self.hh < other.hh
        elif self.mm != other.mm:  # noqa: RET505
            return self.mm < other.mm
        elif self.ss != other.ss:
            return self.ss < other.ss
        elif self.nn != other.nn:
            return self.nn < other.nn
        return False

    def __le__(self, other: object) -> bool:
        """
        Compares the Timestamp object with another Timestamp object to determine if it is less than or equal to.

        Args:
            other: The Timestamp object to compare with.

        Returns:
            bool: True if the Timestamp object is less than or equal to the other Timestamp object, False otherwise.
        """
        if not isinstance(other, Timestamp):
            return NotImplemented
        if self.hh != other.hh:
            return self.hh <= other.hh
        elif self.mm != other.mm:  # noqa: RET505
            return self.mm <= other.mm
        elif self.ss != other.ss:
            return self.ss <= other.ss
        elif self.nn != other.nn:
            return self.nn <= other.nn
        return True

    def __gt__(self, other: object) -> bool:
        """
        Compares the Timestamp object with another Timestamp object to determine if it is greater than.

        Args:
            other: The Timestamp object to compare with.

        Returns:
            bool: True if the Timestamp object is greater than the other Timestamp object, False otherwise.
        """
        if not isinstance(other, Timestamp):
            return NotImplemented
        if self.hh != other.hh:
            return self.hh > other.hh
        elif self.mm != other.mm:  # noqa: RET505
            return self.mm > other.mm
        elif self.ss != other.ss:
            return self.ss > other.ss
        elif self.nn != other.nn:
            return self.nn > other.nn
        return False

    def __ge__(self, other: object) -> bool:
        """
        Compares the Timestamp object with another Timestamp object to determine if it is greater than or equal to.

        Args:
            other: The Timestamp object to compare with.

        Returns:
            bool: True if the Timestamp object is greater than or equal to the other Timestamp object, False otherwise.
        """
        if not isinstance(other, Timestamp):
            return NotImplemented
        if self.hh != other.hh:
            return self.hh >= other.hh
        elif self.mm != other.mm:  # noqa: RET505
            return self.mm >= other.mm
        elif self.ss != other.ss:
            return self.ss >= other.ss
        elif self.nn != other.nn:
            return self.nn >= other.nn
        return True

    def __str__(self) -> str:
        return self.ts

    def __getitem__(self, index: int) -> int:
        """
        Retrieves the element at the specified index from the Timestamp object.

        Args:
            index: The index of the element to retrieve.

        Returns:
            int: The element at the specified index.
        """
        return (self.hh, self.mm, self.ss, self.ss)[index]

    @property
    def ts(self) -> str:
        """
        Generates and returns the timestamp string specified by the object.

        Returns:
            str: A formatted timestamp string based on the object's properties and format.

        The timestamp format is determined by the 'form' attribute and includes hours (HH),
        minutes (MM), seconds (SS), and nanoseconds (NN) as applicable. The method constructs
        the timestamp string according to the specified format, including or omitting parts
        based on the format and the presence of non-zero values.
        """
        format_match = re.match(
            r"^(([Hh]{1,2}):)?([Mm]{1,2}):([Ss]{1,2})(\.([Nn]{1,9}))?$",
            self.form,
        )
        assert format_match is not None
        format_groups = format_match.groups()
        timestamp_format = [format_groups[i] is not None for i in (1, 2, 3, 5)]

        timestamp_string = ""
        if timestamp_format[0] or self._hh:
            timestamp_string += f"{self.hh:0=2d}:"
        if timestamp_format[1] or self._mm:
            timestamp_string += f"{self.mm:0=2d}:"
        if timestamp_format[2] or self._ss:
            timestamp_string += f"{self.ss:0=2d}"
        if timestamp_format[3] or self._nn:
            timestamp_string += f"{self.nn / 1000000000:.9f}".rstrip("0")[1:] if self.nn else ".0"
        return timestamp_string

    @ts.setter
    def ts(self, timestamp: Timestamp) -> None:
        """
        Set a new timestamp.

        timestamp (str, int):
            A str of a timestamp acceptable to mkvmerge or an int representing seconds. This value will be
            the basis of the timestamp.
        """
        if not isinstance(timestamp, (int, str)):
            msg = f'"{type(timestamp)}" is not str or int type'
            raise TypeError(msg)
        self._hh = None
        self._mm = None
        self._ss = None
        self._nn = None
        self.extract(timestamp)

    @property
    def hh(self) -> int:
        """
        Get the hours component of the timestamp.

        Returns:
            int: The hours value.
        """
        return cast(int, self._hh)

    @hh.setter
    def hh(self, value: int) -> None:
        """
        Set the hours component of the timestamp.

        Args:
            value (int): The new hours value.
        """
        self._hh = value

    @property
    def mm(self) -> int:
        """
        Get the minutes component of the timestamp.

        Returns:
            int: The minutes value.
        """
        return cast(int, self._mm)

    @mm.setter
    def mm(self, value: int) -> None:
        """
        Set the minutes component of the timestamp.

        Args:
            value (int): The new minutes value. If greater than or equal to 60, it will be set to 0.
        """
        self._mm = value if value < 60 else 0  # noqa: PLR2004

    @property
    def ss(self) -> int:
        """
        Get the seconds component of the timestamp.

        Returns:
            int: The seconds value.
        """
        return cast(int, self._ss)

    @ss.setter
    def ss(self, value: int) -> None:
        """
        Set the seconds component of the timestamp.

        Args:
            value (int): The new seconds value. If greater than or equal to 60, it will be set to 0.
        """
        self._ss = value if value < 60 else 0  # noqa: PLR2004

    @property
    def nn(self) -> int:
        """
        Get the nanoseconds component of the timestamp.

        Returns:
            int: The nanoseconds value.
        """
        return cast(int, self._nn)

    @nn.setter
    def nn(self, value: int) -> None:
        """
        Set the nanoseconds component of the timestamp.

        Args:
            value (int): The new nanoseconds value. If greater than or equal to 1,000,000,000, it will be set to 0.
        """
        self._nn = value if value < 1000000000 else 0  # noqa: PLR2004

    @property
    def form(self) -> str:
        """
        Get the format string for the timestamp.

        Returns:
            str: The format string.
        """
        return self._form

    @form.setter
    def form(self, form: str) -> None:
        """
        Set the format string for the timestamp.

        Args:
            form (str): The new format string.
        """
        self._form = form

    @staticmethod
    def verify(timestamp: str) -> bool:
        """
        Verify a timestamp has the proper form to be used in mkvmerge.

        timestamp (str):
            The timestamp to be verified.
        """
        if not isinstance(timestamp, str):
            msg = f'"{type(timestamp)}" is not str type'
            raise TypeError(msg)
        return bool(_TIMESTAMP_PATTERN.match(timestamp))

    def extract(self, timestamp: str | int) -> None:
        """
        Extracts time info from a timestamp.

        timestamp (str, int):
            A str of a timestamp acceptable to mkvmerge or an int representing seconds. The timing info will be
            extracted from this parameter.
        """
        if not isinstance(timestamp, (str, int)):
            msg = f'"{type(timestamp)}" is not str or int type'
            raise TypeError(msg)
        if isinstance(timestamp, str) and not Timestamp.verify(
            timestamp,
        ):
            msg = f'"{timestamp}" is not a valid timestamp'
            raise ValueError(msg)
        if isinstance(timestamp, str):
            self.splitting_timestamp(timestamp)
        elif isinstance(timestamp, int):
            self._hh = int(timestamp / 3600)
            self._mm = int(timestamp % 3600 / 60)
            self._ss = int(timestamp % 3600 % 60)
            self._nn = 0

    def splitting_timestamp(self, timestamp: str) -> None:
        """
        This method splits the given timestamp string into hours, minutes, seconds, and nanoseconds. The timestamp
        string should be in the format "HH:MM:SS.NNNNNNNNN", where HH represents hours,
        MM represents minutes, SS represents seconds, and NNNNNNNNN represents nanoseconds.

        Parameters
        ----------
        timestamp : str
            The timestamp string to be split.

        Returns
        -------
        None

        Example
        -------
        >>> obj = Timestamp("12:34:56.789012345")
        >>> print(obj)
        12:34:56.789012345
        >>> assert obj.hh == 12
        >>> assert obj.mm == 34
        >>> assert obj.ss == 56
        >>> assert obj.nn == 789012345

        # Assuming self._hh, self._mm, self._ss, and self._nn are all None initially
        The hours (self.hh) will be set to 12.
        The minutes (self.mm) will be set to 34.
        The seconds (self.ss) will be set to 56.
        The nanoseconds (self.nn) will be set to 789012345.
        """
        timestamp_match = re.match(
            r"^(([0-9]{1,2}):)?([0-9]{1,2}):([0-9]{1,2})(\.([0-9]{1,9}))?$",
            timestamp,
        )
        assert timestamp_match is not None
        timestamp_groups = timestamp_match.groups()

        timestamp_lst = [cast(str, timestamp_groups[i]) for i in (1, 2, 3, 4)]
        timestamp_clean = []
        for ts in timestamp_lst:
            if ts is None:
                timestamp_clean.append(0)
            else:
                timestamp_clean.append(float(ts))
        self.hh = int(timestamp_clean[0]) if self._hh is None else self._hh
        self.mm = int(timestamp_clean[1]) if self._mm is None else self._mm
        self.ss = int(timestamp_clean[2]) if self._ss is None else self._ss
        self.nn = int(timestamp_clean[3] * 1000000000) if self._nn is None else self._nn


if __name__ == "__main__":
    pass
